# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

from functools import partial
from typing import Callable, List, NamedTuple

import jax.numpy as jnp
import numpy as np
from jax.scipy.stats import multivariate_normal
from scipy.stats import multivariate_t


def _rand_gaussian_mixture(n: int, pi: list, mu, cov):
    x = []
    for _ in range(n):
        z_i = np.argmax(np.random.multinomial(1, pi))
        x_i = np.random.multivariate_normal(mu[z_i], cov[z_i], size=1).T
        x.append(x_i)
    return jnp.array(x).squeeze()


def _rand_gaussian_banana(mu: np.ndarray, cov: np.ndarray):
    d = len(mu)
    b = 0.025
    z = np.random.multivariate_normal(np.zeros(d), np.diag(cov), size=1).T
    x = np.zeros_like(z)
    x = z
    x[1] += b * z[0] ** 2 - 100 * b
    y = x.squeeze() + mu
    return y


def _rand_gaussian_banana_mixture(
    mu: np.ndarray, pi: np.ndarray, cov_diag: np.ndarray, size: int = 10_000
):
    _, d = mu.shape
    x = np.zeros((size, d))
    for i in range(size):
        zi = np.argmax(np.random.multinomial(1, pi))
        x[i] = _rand_gaussian_banana(mu[zi], cov_diag)
    return x


def _rand_t_banana(mu: np.ndarray, cov: np.ndarray, df: int):
    d = len(mu)
    b = 0.025
    z = multivariate_t.rvs(np.zeros(d), cov, df=df, size=1)
    x = np.zeros_like(z)
    x = z
    x[1] += b * z[0] ** 2 - 100 * b
    return x + mu


def _rand_t_banana_mixture(
    mu: np.ndarray, pi: np.ndarray, cov_diag: np.ndarray, df: int, size: int = 10_000
):
    _, d = mu.shape
    x = np.zeros((size, d))
    for i in range(size):
        zi = np.argmax(np.random.multinomial(1, pi))
        x[i] = _rand_t_banana(mu[zi], np.diag(cov_diag), df)
    return x


def _logpdf_multivariate_t(x, loc, df, prec_U, dim):
    dev = x - loc
    maha = jnp.sum(jnp.square(dev * prec_U))
    t = 0.5 * (df + dim)
    logpdf = -t * jnp.log(1.0 + (1.0 / df) * maha) + 2 * dim
    return logpdf


def _pdf_multivariate_t(x, loc, df, prec_U, dim):
    log_pdf = _logpdf_multivariate_t(x, loc, df, prec_U, dim)
    return jnp.exp(log_pdf)


class Distribution(NamedTuple):
    d: int
    logprob_fn: Callable
    rand: Callable


class GaussianMixture:
    def __new__(
        cls, d: int, pi: List[int], mu: jnp.ndarray, cov: jnp.ndarray
    ) -> Distribution:
        def logprob_fn(x: jnp.ndarray):
            K = 1.5 * d
            pdf = pi[0] * jnp.exp(K + multivariate_normal.logpdf(x, mu[0], cov[0]))
            for i in range(1, len(pi)):
                pdf += pi[i] * jnp.exp(K + multivariate_normal.logpdf(x, mu[i], cov[i]))
            return jnp.log(pdf)

        rand = partial(_rand_gaussian_mixture, pi=pi, mu=mu, cov=cov)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)


class gBananaMixture:
    """Gaussian banana mixture."""

    def __new__(cls, d: int, pi: List[int], mu: List[np.ndarray]) -> Distribution:
        dim = d
        cov_diag = np.array([30] + [1] * (d - 1))

        def logprob_fn(x: np.ndarray):
            b = 0.025
            z = x - mu[0]
            z = z.at[1].set(x[1] - b * z[0] ** 2 + 100 * b - mu[0][1])
            pdf = pi[0] * multivariate_normal.pdf(z, jnp.zeros(dim), jnp.diag(cov_diag))
            for i in range(1, len(pi)):
                z = x - mu[i]
                z = z.at[1].set(x[1] - b * z[0] ** 2 + 100 * b - mu[i][1])
                pdf += pi[i] * multivariate_normal.pdf(
                    z, jnp.zeros(dim), jnp.diag(cov_diag)
                )
            return jnp.log(pdf)

        rand = partial(_rand_gaussian_banana_mixture, mu=mu, pi=pi, cov_diag=cov_diag)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)


class tBananaMixture:
    """Student-t banana mixture."""

    def __new__(
        cls, d: int, pi: List[int], mu: List[np.ndarray], df: int
    ) -> Distribution:
        dim = d
        cov_diag = np.array([30] + [1] * (d - 1))
        prec_U = jnp.sqrt(1.0 / cov_diag)

        def logprob_fn(x: np.ndarray):
            b = 0.025
            z = x - mu[0]
            z = z.at[1].set(x[1] - b * z[0] ** 2 + 100 * b - mu[0][1])
            pdf = pi[0] * _pdf_multivariate_t(z, jnp.zeros(dim), df, prec_U, dim)
            for i in range(1, len(pi)):
                z = x - mu[i]
                z = z.at[1].set(x[1] - b * z[0] ** 2 + 100 * b - mu[i][1])
                pdf += pi[i] * _pdf_multivariate_t(z, jnp.zeros(dim), df, prec_U, dim)
            return jnp.log(pdf)

        rand = partial(_rand_t_banana_mixture, mu=mu, pi=pi, cov_diag=cov_diag, df=df)
        return Distribution(d=d, logprob_fn=logprob_fn, rand=rand)



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

from typing import Callable

import jax
import jax.numpy as jnp

from kernax.types import Array


class KernelQuantization:
    """Implements kernel quantization by greedily minimizing the maximum mean discrepancy.

    [1] Chen, Y., Welling, M., & Smola, A. (2012).
    Super-samples from kernel herding. arXiv preprint arXiv:1203.3472

    [2] Teymur, O., Gorham, J., Riabiz, M., & Oates, C. (2021, March).
    Optimal quantisation of probability measures using maximum mean discrepancy. In International Conference on Artificial Intelligence and Statistics (pp. 1027-1035). PMLR.

    Examples:
    --------
    Assume that we want to summarize a two-dimensional Gaussian sample of size 10,000.

    .. code::

        import jax
        import jax.random as jr

        key = jr.PRNGKey(0)
        X = jr.normal(key, (10_000, 2))

    Now that we have a toy sample, let's pick a kernel function.
    Here, we use the energy kernel function. In this case, the maximum mean discrepancy reduces to the energy distance.

    .. code::

        from kernax.kernels import Energy
        from kernax import KernelQuantization

        quant_fn = KernelQuantization(X, kernel_fn=Energy)
        idx = quant_fn(m = 1_000)

    The output `idx` gathers the selected indices. Let's plot the result.

    .. code::

        import matplotlib.pyplot as plt

        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(X[:, 0], X[:, 1], ls="", marker="o", color="k", label="Initial sample")
        ax.plot(X[idx, 0], X[idx, 1], ls="", marker="o", color="r", label="Selected sample")
        ax.legend(fontsize=14)

    """

    def __new__(
        cls,
        x: Array,
        kernel_fn: Callable,
    ) -> Callable:
        def gram_matrix_fn(X: Array, Y: Array):
            return jax.vmap(lambda xx: jax.vmap(lambda yy: kernel_fn(xx, yy))(Y))(X)

        k0_diag = jax.vmap(lambda xx: kernel_fn(xx, xx))
        kmap = jax.jit(jax.vmap(kernel_fn, (None, 0)))

        def mmd_quantization_fn(m: int) -> Array:
            Kmat = gram_matrix_fn(x, x)
            k0_mean = jnp.mean(Kmat, axis=1)
            k0 = k0_diag(x)
            obj = k0 - 2.0 * k0_mean
            init = jnp.argmin(obj)

            def thinning_step_fn(carry, xs):
                idx, obj = carry
                ki = kmap(x[idx], x)
                obj = obj + 2.0 * ki - 2.0 * k0_mean
                new_idx = jnp.argmin(obj)
                return (new_idx, obj), new_idx

            _, idx = jax.lax.scan(thinning_step_fn, (init, obj), jnp.arange(1, m))
            return jnp.append(init, idx)

        return mmd_quantization_fn



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

from typing import Callable

import jax
import jax.numpy as jnp

from kernax.kernels import IMQ, Energy, GetSteinFn
from kernax.types import Array
from kernax.utils import median_heuristic


def MMD(x: Array, y: Array) -> float:
    r"""Implements the V-estimator the maximum mean discrepancy.

    Maximum mean discrepancy between two probability distributions :math:`P` and :math:`Q` is defined as:

    .. math::

        \mathrm{MMD}^2(P,Q) = \mathbb{E}_{x\sim P}\mathbb{E}_{y\sim P}k(x,y) + \mathbb{E}_{x\sim Q}\mathbb{E}_{y\sim Q}k(x,y) - 2 \mathbb{E}_{x\sim P}\mathbb{E}_{y\sim Q}k(x,y)

    Given two samples :math:`\{x_i\}_{i=1}^{n}` and :math:`\{y_i\}_{i=1}^{m}`, this function estimates MMD with a V-statistics.

    Examples:
    --------
    Assume that we have two samples from Gaussian distributions.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        rng_key1, rng_key2 = jax.random.split(key)

        x = jax.random.normal(rng_key1, (10_000,2))
        y = jax.random.normal(rng_key2, (10_000,2)) + jnp.array([0.1, 0.5])[:,None]


    The maximum mean discrepancy estimated by a V-statistics can be obtained as follows.

    .. code::

        mmd = MMD(x, y)

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    y
        Sample matrix of size :math:`(n, d)`

    Returns:
    -------
    The V-estimator of the maximum mean discrepancy between the provided samples.
    """
    kxx = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(x))(x)
    kyy = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(y))(y)
    kxy = jax.vmap(lambda x1: jax.vmap(lambda y1: Energy(x1, y1))(x))(y)
    mmd = (
        (1.0 / x.shape[0] ** 2) * kxx.sum()
        + (1.0 / y.shape[0] ** 2) * kyy.sum()
        - (2.0 / (x.shape[0] * y.shape[0])) * kxy.sum()
    )
    return jnp.sqrt(mmd)


def KSD(x: Array, sx: Array, kernel_fn: Callable = None) -> float:
    r"""Implements the V-estimator of kernelized Stein discrepancy.

    Kernelized Stein discrepancy between two probability distributions is defined as:

    .. math::
        \mathrm{KSD}^2(P,Q) = E_{x\sim Q}E_{y\sim Q} k_p(x,y)

    where :math:`k_p` denotes the Stein kernel.

    Examples:
    --------
    Assume that we have a sample from a Gaussian distribution.

    .. code::

        import jax

        key = jax.random.PRNGKey(0)
        x = jax.random.normal(key, (10_000,2)) + jnp.array([1.0, 1.0])[None,:]

    The target distrubution is also a Gaussian distribution but centered. Hence, the scores are obtained as follows.

    .. code::

        score_fn = lambda x: -x
        sx = score_fn(x)

    We can now compute KSD.

    .. code::

        ksd = KSD(x, sx)

    Note that we can also specify the underlying kernel involved in the Stein kernel.

    This function uses the inverse multi-quadratric kernel by default but any other kernel function can be passed as an argument.

    .. code::

        from kernax.kernels import Gaussian

        ksd = KSD(x, sx, Gaussian)

    The lengthscale (or bandwidth) of the underyling `kernel_fn` is automatically set to the median heuristic.

    Parameters
    ----------

    x
        Sample matrix of size :math:`(n, d)`
    sx
        Matrix of size :math:`(n, d)` containing the scores of the sample :math:`x`.
    kernel_fn
        Callable corresponding the the underlying kernel function in the Stein kernel (Default: IMQ kernel).

    Returns:
    -------
    The V-estimator of kernelized Stein discrepancy given the provided sample and chosen score function.
    """
    if kernel_fn is None:
        lengthscale = jnp.array(median_heuristic(x))
        kernel_fn = jax.tree_util.Partial(IMQ, lengthscale=lengthscale)

    kp_fn = GetSteinFn(kernel_fn)
    kp = jax.vmap(lambda a, b: jax.vmap(lambda c, d: kp_fn(a, b, c, d))(x, sx))(x, sx)
    ksd = jnp.sqrt(jnp.sum(kp)) / x.shape[0]
    return ksd



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

import jax

#: JAX Array
Array = jax.Array

#: JAX PRNGKey
PRNGKey = jax.Array



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

from typing import Callable

import jax
import jax.numpy as jnp

from kernax.types import Array


def IMQ(x: Array, y: Array, lengthscale: float):
    """Inverse multi-quadratric kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    lengthscale
        Scalar lengthscale / bandwidth

    Returns:
    -------
    Value of the kernel function k(x,y)
    """
    return 1.0 / jnp.sqrt(1.0 + jnp.sum((x - y) ** 2) / lengthscale**2)


def Gaussian(x: Array, y: Array, lengthscale: float):
    """Gaussian kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    lengthscale
        Scalar lengthscale / bandwidth

    Returns:
    -------
    Value of the kernel function k(x,y)
    """
    return jnp.exp(-0.5 * jnp.sum((x - y) ** 2) / lengthscale**2)


def Energy(x: Array, y: Array):
    """Distance induced kernel function.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`

    Returns:
    -------
    Value of the kernel function k(x,y)
    """
    # return jnp.sqrt(jnp.sum(x**2)) + jnp.sqrt(jnp.sum(y**2)) - jnp.sqrt(jnp.sum((x-y)**2))
    return (
        jnp.sqrt(x @ x)
        + jnp.sqrt(y @ y)
        - jnp.sqrt(jnp.clip(x @ x + y @ y - 2 * x @ y, a_min=0))
    )


def SteinIMQ(x: Array, sx: Array, y: Array, sy: Array, lengthscale: float):
    """Langevin Stein kernel with the IMQ as the underlying kernel.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    sx
        Score functon evaluated at x, vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    sy
        Score function evaluetaed at y, vector of dimension :math:`d`

    Returns:
    -------
    Value the Stein IMQ kernel :math:`k_p(x,y)`
    """
    d = len(x)
    sqdist = jnp.sum((x - y) ** 2)
    qf = 1.0 / (1.0 + sqdist / lengthscale**2)
    t3 = jnp.dot(sx, sy) * jnp.sqrt(qf)
    t2 = (1.0 / lengthscale**2) * (d + jnp.dot(sx - sy, x - y)) * qf ** (3 / 2)
    t1 = (-3.0 / lengthscale**4) * sqdist * qf ** (5 / 2)
    return t1 + t2 + t3


def SteinGaussian(x: Array, sx: Array, y: Array, sy: Array, lengthscale: float):
    """Langevin Stein kernel with the Gaussian kernel as the underlying kernel.

    Parameters
    ----------

    x
        Vector of dimension :math:`d`
    sx
        Score functon evaluated at x, vector of dimension :math:`d`
    y
        Vector of dimension :math:`d`
    sy
        Score function evaluetaed at y, vector of dimension :math:`d`

    Returns:
    -------
    Value the Stein Gaussian kernel :math:`k_p(x,y)`
    """
    d = len(x)
    sqdist = jnp.sum((x - y) ** 2)
    kxy = jnp.exp(-0.5 * sqdist / lengthscale**2)
    t1 = d / lengthscale**2 - sqdist / lengthscale**4
    t2 = jnp.dot(sx - sy, x - y) / lengthscale**2
    t3 = jnp.dot(sx, sy)
    return kxy * (t1 + t2 + t3)


def GetSteinFn(kernel_fn: Callable):
    r"""Helper that builds the Stein kernel function :math:`k_p(x,y)` given an arbitrary underlying kernel :math:`k(x,y)`.

    Parameters
    ----------

    kernel_fn
        Callable kernel function of the form :math:`(x,y) \mapsto k(x,y)`. Any hyparameters such as the lengthscale should be fixed with `jax.tree_util.Partial`.

    Returns:
    -------
    Callable function of the form :math:`(x, s_x, y, s_y) \mapsto k_p(x, s_x, y, s_y)` where :math:`x` and :math:`y` are the vectors of observations and :math:`s_x` and :math:`s_y` are the associated scores.
    """
    grad1_fn = jax.grad(kernel_fn, argnums=0)
    grad2_fn = jax.grad(kernel_fn, argnums=1)
    hessian_fn = jax.jacobian(grad2_fn, argnums=0)

    def kp_fn(x, sx, y, sy):
        return (
            jnp.dot(sx, sy) * kernel_fn(x, y)
            + jnp.dot(sx, grad2_fn(x, y))
            + jnp.dot(sy, grad1_fn(x, y))
            + jnp.trace(hessian_fn(x, y))
        )

    return kp_fn



# This file is subject to the terms and conditions defined in
# file 'LICENSE.txt', which is part of this source code package.

from .bjsamplers import hmc, mala, nuts, rmh
from .discrepancies import KSD, MMD
from .quantization import (
    KernelQuantization,
)
from .thinning import RegularizedSteinThinning, SteinThinning
from .utils import laplace_log_p_hardplus, laplace_log_p_softplus

__all__ = [
    "SteinThinning",
    "RegularizedSteinThinning",
    "KernelQuantization",
    "laplace_log_p_hardplus",
    "laplace_log_p_softplus",
    "MMD",
    "KSD",
    "rmh",
    "hmc",
    "nuts",
    "mala",
]
